#########################
#########################
###                   ###
###  MUTE             ###
###  William Woodley  ###
###  15 July 2022     ###
###                   ###
#########################
#########################

# Import packages

import os

import numpy as np
import scipy.integrate as scii
import scipy.interpolate as sciint

import mute.constants as constants
import mute.surface as surface
import mute.propagation as propagation

# Perform the convolution for underground fluxes


def _do_convolution(s_fluxes, survival_probability_tensor):

    """This function performs the convolution to calculate underground fluxes given a surface flux matrix and survival probability tensor."""

    # Check values

    constants._check_constants()

    s_fluxes = np.atleast_2d(s_fluxes)

    assert s_fluxes.shape[0] == len(
        constants.ENERGIES
    ), f"The surface flux matrix must use the default MUTE energy grid, which has length {len(constants.ENERGIES)}, not {s_fluxes.shape[0]}."

    # In order for the dot product in the convolution to work, move the axes
    # Move axis 0 (surface energies) to position 2
    # Original: (i, x, u)
    # Moved:    (x, u, i)

    survival_probability_tensor = np.moveaxis(survival_probability_tensor, 0, 2)

    # Construct a matrix of energy bin widths

    widths = np.repeat(constants._E_WIDTHS, s_fluxes.shape[1]).reshape(
        (len(constants.ENERGIES), s_fluxes.shape[1])
    )

    # Calculate the underground flux tensor

    u_fluxes_from_convolution = (
        survival_probability_tensor.dot((s_fluxes.T * constants._E_WIDTHS).T) / widths
    )

    return u_fluxes_from_convolution


# Calculate underground fluxes


def calc_u_fluxes(
    s_fluxes=None,
    survival_probability_tensor=None,
    full_tensor=False,
    primary_model="GSF",
    interaction_model="SIBYLL-2.3c",
    atmosphere="CORSIKA",
    location="USStd",
    month=None,
    output=None,
    file_name="",
    force=False,
):

    """
    Calculate underground fluxes in units of [(cm^2 s sr MeV)^-1] for default surface energy grid and zenith angles.

    The default surface energy grid is given by constants.ENERGIES, and the default zenith angles are given by constants._ANGLES.

    Parameters
    ----------
    s_fluxes : NumPy ndarray, optional (default: taken from surface.load_s_fluxes_from_file())
        A surface flux matrix of shape (91, 20).

    survival_probability_tensor : NumPy ndarray, optional (default: taken from propagation.load_survival_probability_tensor_from_file())
        A survival probability tensor of shape (91, 28, 91).

    full_tensor : bool, optional (default: False)
        If True, the full tensor of shape (28, 91, 20) will be returned. Otherwise, if the overburden type is flat, a two-dimensional matrix of shape (91, 28) will be returned.

    primary_model : str in {"GSF", "HG", "GH", "ZS", "ZSP", "PL27"} or tuple, optional (default: "GSF")
        The primary flux model to use in MCEq. Options:
        GSF  = GlobalSplineFitBeta
        HG   = HillasGaisser2012 (H3a)
        GH   = GaisserHonda
        ZS   = Zatsepin-Sokolskaya (Default)
        ZSP  = Zatsepin-Sokolskaya (PAMELA)
        PL27 = SimplePowerlaw27
        Alternatively, this can be set with a tuple. For example: (pm.GaisserStanevTilav, "3-gen").

    interaction_model : str, optional (default: "SIBYLL-2.3c")
        The hadronic interaction model to use in MCEq. See the Tutorial or MCEq documentation for a list of options.

    atmosphere : {"CORSIKA", "MSIS00"}, optional (default: "CORSIKA")
        The atmospheric model to use in MCEq. For US Standard Atmosphere, use "CORSIKA". For seasonal variations, use "MSIS00".

    location : str, optional (default: "USStd")
        The name of the location for which to calculate the surface fluxes. See the Tutorial or MCEq documentation for a list of options.

    month : str, optional (default: None)
        The month for which to calculate the surface fluxes. For US Standard Atmosphere, use None. For seasonal variations, use the month name.

    output : bool, optional (default: taken from constants.get_output())
        If True, an output file will be created to store the results.

    file_name : str, optional (default: constructed from input parameters)
        The name of the file in which to store the results. If output is False or None, this is ignored.

    force : bool
        If True, force the calculation of new matrices and the creation of new directories if required.

    Returns
    -------
    u_fluxes : NumPy ndarray
        A multi-dimensional array containing the underground fluxes. For flat overburdens, the shape will be (91, 28) for the default zenith angles, and the fluxes will be in units of [(cm^2 s sr MeV)^-1]. For mountain overburdens (or if full_tensor is True), the shape will be (28, 91, 20), and the fluxes will be in units of [(cm^2 s sr MeV^2)^-1].
    """

    # Check values

    constants._check_constants()

    if output is None:
        output = constants.get_output()

    # Get the surface flux matrix and survival probability tensor
    # If the calculation functions have to be run through the getter functions, the global output setting is used

    if constants.get_verbose() > 1:
        print("Calculating underground fluxes.")

    if s_fluxes is None:

        s_fluxes = surface.load_s_fluxes_from_file(
            primary_model, interaction_model, atmosphere, location, month, force=force
        )

    if survival_probability_tensor is None:

        survival_probability_tensor = (
            propagation.load_survival_probability_tensor_from_file(force=force)
        )

    # Check that the surface flux matrix and survival probability tensor have been loaded properly

    if s_fluxes is None or survival_probability_tensor is None:

        raise Exception(
            "Underground fluxes not calculated. The surface flux matrix or survival probability tensor was not provided or loaded correctly."
        )

    assert (
        s_fluxes.shape[0] == survival_probability_tensor.shape[2]
    ), f"The surface flux matrix and survival probability tensor must both use the default MUTE energy grid, which has length {len(constants.ENERGIES)}, not {s_fluxes.shape[0]} or {survival_probability_tensor.shape[2]}."

    # Perform the convolution

    u_fluxes_from_convolution = _do_convolution(s_fluxes, survival_probability_tensor)

    # Return an interpolated matrix for flat overburdens

    if constants.get_overburden() == "flat" and not full_tensor:

        # Interpolate the u_fluxes tensor to a constant grid
        # This allows the np.diagonal() function to be used to calculate the matrix of relevant fluxes from the full tensor

        interp_u_fluxes_from_convolution = sciint.interp1d(
            constants.ANGLES_FOR_S_FLUXES,
            u_fluxes_from_convolution,
            axis=2,
            kind="linear",
            fill_value=0,
        )(constants._ANGLES)

        # Extract the (u_energies, angles) diagonal from the convolution tensor

        u_fluxes = np.diagonal(interp_u_fluxes_from_convolution, axis1=0, axis2=2)

        # Slice the angles corresponding to constants.angles

        u_fluxes = u_fluxes[:, (len(constants._ANGLES) - len(constants.angles)) :]

        if constants.get_verbose() > 1:
            print("Finished calculating underground fluxes.")

        # Write the results to the file

        if output:

            constants._check_directory(
                os.path.join(constants.get_directory(), "underground"), force=force
            )

            if file_name == "" or not isinstance(file_name, str):

                file_name = os.path.join(
                    constants.get_directory(),
                    "underground",
                    "{0}_Underground_Fluxes.txt".format(constants.get_lab()),
                )

            file_out = open(file_name, "w")

            for u in range(len(constants.ENERGIES)):

                for j in range(len(constants.angles)):

                    file_out.write(
                        "{0:1.14f} {1:1.5f} {2:1.14e}\n".format(
                            constants.ENERGIES[u], constants.angles[j], u_fluxes[u, j]
                        )
                    )

            file_out.close()

            if constants.get_verbose() > 1:
                print(f"Underground fluxes written to {file_name}.")

        return u_fluxes

    # Return the whole tensor for mountains

    elif constants.get_overburden() == "mountain" or full_tensor:

        u_fluxes = np.squeeze(u_fluxes_from_convolution)

        if constants.get_verbose() > 1:
            print("Finished calculating underground fluxes.")

        # Write the results to the file

        if output:

            constants._check_directory(
                os.path.join(constants.get_directory(), "underground"), force=force
            )

            if file_name == "" or not isinstance(file_name, str):

                file_name = os.path.join(
                    constants.get_directory(),
                    "underground",
                    "{0}_Underground_Fluxes.txt".format(constants.get_lab()),
                )

            file_out = open(file_name, "w")

            for x in range(len(constants._SLANT_DEPTHS)):

                for u in range(len(constants.ENERGIES)):

                    for j in range(len(constants.ANGLES_FOR_S_FLUXES)):

                        file_out.write(
                            "{0:1.5f} {1:1.14f} {2:1.5f} {3:1.14e}\n".format(
                                constants._SLANT_DEPTHS[x],
                                constants.ENERGIES[u],
                                constants.ANGLES_FOR_S_FLUXES[j],
                                u_fluxes[x, u, j],
                            )
                        )

            file_out.close()

            if constants.get_verbose() > 1:
                print(f"Underground fluxes written to {file_name}.")

        return u_fluxes_from_convolution

    else:

        raise NotImplementedError(
            'Overburdens of type {0} are not available. The only options are "flat" and "mountain".'.format(
                constants.get_overburden()
            )
        )


# Create the interpolator to calculate double-differential underground intensities


def _create_interpolator(u_intensities):

    """This function creates the interpolator object to use when computing double-differential underground intensities."""

    global _interpolator

    _interpolator = sciint.interp2d(
        constants._SLANT_DEPTHS,
        constants.ANGLES_FOR_S_FLUXES,
        np.log(u_intensities.T),
        kind="linear",
        fill_value=-np.inf,
    )

    return _interpolator


# Calculate the underground intensities


def calc_u_intensities(method, output=None, file_name="", force=False, **kwargs):

    """
    Calculate underground intensities in units of [(cm^2 s sr)^-1].

    Parameters
    ----------
    method : str
        The type of underground intensities to calculate. Options:
        sd = Single-differential underground intensities for flat overburdens
        eq = Vertical-equivalent underground intensities for flat overburdens
        tr = True vertical underground intensities for flat overburdens
        dd = Double-differential underground intensities for mountain overburdens

    output : bool, optional (default: taken from constants.get_output())
        If True, an output file will be created to store the results.

    file_name : str, optional (default: constructed from input parameters)
        The name of the file in which to store the results. If output is False or None, this is ignored.

    force : bool
        If True, force the calculation of new matrices and the creation of new directories if required.

    Other Parameters
    -----------------
    u_fluxes : NumPy ndarray, optional (default: calcaulated from default surface fluxes and survival probabilities)
        An underground flux tensor of shape (28, 91, 20).

    s_fluxes : NumPy ndarray, optional (default: taken from surface.load_s_fluxes_from_file())
        A surface flux matrix of shape (91, 20).

    survival_probability_tensor : NumPy ndarray, optional (default: taken from propagation.load_survival_probability_tensor_from_file())
        A survival probability tensor of shape (91, 28, 91).

    angles : array-like, optional (default: taken from consants.angles)
        An array of zenith angles in [degrees] to calculate the underground intensities for. These are only used if method is "sd" or "eq".

    depths : array-like, optional (default: taken from consants.slant_depths)
        An array of slant depths in [km.w.e.] to calculate the underground intensities for. These are only used if method is "tr".

    E_th : float, optional (default: 0)
        An energy threshold in [MeV].

    primary_model : str in {"GSF", "HG", "GH", "ZS", "ZSP", "PL27"} or tuple, optional (default: "GSF")
        The primary flux model to use in MCEq. Options:
        GSF  = GlobalSplineFitBeta
        HG   = HillasGaisser2012 (H3a)
        GH   = GaisserHonda
        ZS   = Zatsepin-Sokolskaya (Default)
        ZSP  = Zatsepin-Sokolskaya (PAMELA)
        PL27 = SimplePowerlaw27
        Alternatively, this can be set with a tuple. For example: (pm.GaisserStanevTilav, "3-gen").

    interaction_model : str, optional (default: "SIBYLL-2.3c")
        The hadronic interaction model to use in MCEq. See the Tutorial or MCEq documentation for a list of options.

    atmosphere : {"CORSIKA", "MSIS00"}, optional (default: "CORSIKA")
        The atmospheric model to use in MCEq. For US Standard Atmosphere, use "CORSIKA". For seasonal variations, use "MSIS00".

    location : str, optional (default: "USStd")
        The name of the location for which to calculate the surface fluxes. See the Tutorial or MCEq documentation for a list of options.

    month : str, optional (default: None)
        The month for which to calculate the surface fluxes. For US Standard Atmosphere, use None. For seasonal variations, use the month name.

    Returns
    -------
    u_intensities : NumPy ndarray
        An array containing the underground intensities. For flat overburdens (if method is "sd", "eq", or "tr"), the length will be that of angles, and the intensities will be in units of [(cm^2 s sr)^-1]. For mountain overburdens (if method is "dd"), the shape will be (len(constants.mountain.zenith), len(constants.mountain.azimuthal)), and the intensities will be in units of [(cm^2 s sr km.w.e.)^-1].
    """

    # Check values

    constants._check_constants()

    method = method.lower()

    assert method in [
        "sd",
        "eq",
        "tr",
        "dd",
    ], '"{0}" is not a valid method. Use "sd", "tr", "eq", or "dd". See the Tutorial at {1} for an explanation.'.format(
        method,
        "https://github.com/wjwoodley/mute/blob/main/docs/Tutorial.md#calculating-underground-intensities",
    )

    if output is None:
        output = constants.get_output()

    # Pop the keyword arguments

    u_fluxes = kwargs.pop("u_fluxes", None)
    s_fluxes = kwargs.pop("s_fluxes", None)
    survival_probability_tensor = kwargs.pop("survival_probability_tensor", None)
    angles = kwargs.pop("angles", constants.angles)
    depths = kwargs.pop("depths", constants.slant_depths)
    E_th = kwargs.pop("E_th", 0)
    primary_model = kwargs.pop("primary_model", "GSF")
    interaction_model = kwargs.pop("interaction_model", "SIBYLL-2.3c")
    atmosphere = kwargs.pop("atmosphere", "CORSIKA")
    location = kwargs.pop("location", "USStd")
    month = kwargs.pop("month", None)

    if kwargs:

        from warnings import warn as warning

        warning("{0} arguments not used.".format(kwargs))

    # Ensure the arguments are in the proper form
    # Turn the energy threshold value in [MeV] into an index in order to slice the u_fluxes matrix and the energy grid properly when integrating

    angles = np.atleast_1d(angles)
    depths = np.atleast_1d(depths)
    E_th_i = np.argmin(np.abs(constants.ENERGIES - E_th))

    if E_th_i != 0 and constants.get_verbose() > 1:

        print(
            "Setting energy threshold to {0:1.3f} MeV (the closest value in the energy grid to {1} MeV).".format(
                constants.ENERGIES[E_th_i], E_th
            )
        )

    # Calculate the underground fluxes
    # Get the surface flux matrix and survival probability tensor
    # If the calculation functions have to be run through the getter functions, the global output setting is used

    if u_fluxes is None:

        if constants.get_verbose() > 1:
            print("Calculating underground fluxes.")

        if s_fluxes is None:

            s_fluxes = surface.load_s_fluxes_from_file(
                primary_model,
                interaction_model,
                atmosphere,
                location,
                month,
                force=force,
            )

        if survival_probability_tensor is None:

            survival_probability_tensor = (
                propagation.load_survival_probability_tensor_from_file(force=force)
            )

        # Check that the surface flux matrix and survival probability tensor have been loaded properly

        if s_fluxes is None or survival_probability_tensor is None:

            raise Exception(
                "Underground intensities not calculated. The surface flux matrix or survival probability tensor was not provided or loaded correctly."
            )

        assert (
            s_fluxes.shape[0] == survival_probability_tensor.shape[2]
        ), f"The surface flux matrix and survival probability tensor must both use the default MUTE energy grid, which has length {len(constants.ENERGIES)}, not {s_fluxes.shape[0]} or {survival_probability_tensor.shape[2]}."

        # Perform the convolution

        u_fluxes = _do_convolution(s_fluxes, survival_probability_tensor)

        if constants.get_verbose() > 1:
            print("Finished calculating underground fluxes.")

    # Check that the underground flux tensor has the correct shape

    assert u_fluxes.shape == (
        28,
        91,
        20,
    ), f"The underground flux tensor does not have the correct shape. The shape must be (28, 91, 20), not {u_fluxes.shape}. Set full_tensor to True in the underground.calc_u_fluxes() function in order to return a tensor of shape (28, 91, 20). Alternatively, try passing a surface flux matrix and / or a survival probability tensor in as parameters."

    # Calculate the intensities according to the method type
    # sd = Single-differential intensities
    # tr = True vertical intensities
    # eq = Vertical-equivalent intensities
    # dd = Double-differential intensities (for mountains)

    # Calculate the single-differential standard underground intensities or vertical-equivalent underground intensities

    if method == "sd" or method == "eq":

        return _calc_u_intensities_sd(
            method,
            u_fluxes,
            angles,
            E_th_i,
            primary_model,
            interaction_model,
            atmosphere,
            location,
            month,
            output,
            file_name,
            force,
        )

    # Calculate the true vertical underground intensities

    elif method == "tr":

        return _calc_u_intensities_tr(
            u_fluxes,
            depths,
            E_th_i,
            primary_model,
            interaction_model,
            atmosphere,
            location,
            month,
            output,
            file_name,
            force,
        )

    # Calculate the double-differential underground intensities

    elif method == "dd":

        return _calc_u_intensities_dd(
            u_fluxes,
            E_th_i,
            primary_model,
            interaction_model,
            atmosphere,
            location,
            month,
            output,
            file_name,
            force,
        )

    else:

        raise ValueError(
            '"{0}" is not a valid method. Use "sd", "tr", "eq", or "dd". See the Tutorial at {1} for an explanation.'.format(
                method,
                "https://github.com/wjwoodley/mute/blob/main/docs/Tutorial.md#calculating-underground-intensities",
            )
        )


# Calculate single-differential underground intensities


def _calc_u_intensities_sd(
    method,
    u_fluxes,
    angles,
    E_th_i,
    primary_model,
    interaction_model,
    atmosphere,
    location,
    month,
    output,
    file_name,
    force,
):

    """This function calculates single-differential or vertical-equivalent underground intensities for flat overburdens."""

    # Check values

    constants._check_constants()

    if constants.get_overburden() != "flat":

        raise ValueError(
            "Single-differential underground intensities can only be calculated for flat overburdens. See the Tutorial at {0} for an explanation.".format(
                "https://github.com/wjwoodley/mute/blob/main/docs/Tutorial.md#calculating-underground-intensities"
            )
        )

    assert method in ["sd", "eq"]

    if u_fluxes is None:

        raise Exception(
            "Underground intensities not calculated. The underground flux matrix was not calculated correctly."
        )

    # Calculate the underground intensities
    # Zeroth index = Slant depth
    # First index  = Zenith angle (from constants.ANGLES_FOR_S_FLUXES)

    u_intensities = np.zeros(
        (len(constants._SLANT_DEPTHS), len(constants.ANGLES_FOR_S_FLUXES))
    )

    if constants.get_verbose() > 1:
        print("Calculating underground intensities.")

    # Whether provided or calculated, the u_fluxes matrix has to be of the shape (28, 91, 20)
    # Otherwise the cos(theta) approximation results in incorrect intensities

    for x in range(len(constants._SLANT_DEPTHS)):

        for j in range(len(constants.ANGLES_FOR_S_FLUXES)):

            u_intensities[x, j] = scii.simpson(
                u_fluxes[x, E_th_i:, j], constants.ENERGIES[E_th_i:]
            )

    # Interpolate intensities matrix to consistent grid given by set vertical depth
    # This allows the np.diagonal() function to be used to calculate the array of relevant intensities from the full matrix
    # Define a meshgrid for grid depths and angles
    # Define a meshgrid for adjusted depths and angles

    gd, ga = np.meshgrid(constants._SLANT_DEPTHS, constants.ANGLES_FOR_S_FLUXES)
    ad, aa = np.meshgrid(constants.slant_depths, constants.angles)

    u_intensities = np.nan_to_num(
        np.exp(
            sciint.griddata(
                (gd.flatten(), ga.flatten()),
                np.log(u_intensities.T.flatten()),
                (ad, aa),
            )
        )
    )
    u_intensities = np.diagonal(u_intensities)

    # Interpolate intensities to user-input angles
    # If the user has not specified angles, return the whole matrix with no interpolation
    # Check to make sure both arrays have the same number of values, and that those values are equal at each index

    if not (
        len(angles) == len(constants.angles) and np.allclose(angles, constants.angles)
    ):

        u_intensities = np.interp(angles, constants.angles, u_intensities)

    if method == "eq":

        u_intensities = np.cos(np.radians(angles)) * u_intensities

    if constants.get_verbose() > 1:
        print("Finished calculating underground intensities.")

    # Write the results to the file

    if output:

        constants._check_directory(
            os.path.join(constants.get_directory(), "underground"), force=force
        )

        if file_name == "" or not isinstance(file_name, str):

            file_name = os.path.join(
                constants.get_directory(),
                "underground",
                "{0}_Underground_Intensities_{1}.txt".format(
                    constants.get_lab(), method
                ),
            )

        file_out = open(file_name, "w")

        for j in range(len(angles)):

            file_out.write("{0:1.5f} {1:1.14e}\n".format(angles[j], u_intensities[j]))

        file_out.close()

        if constants.get_verbose() > 1:
            print(f"Underground intensities written to {file_name}.")

    return u_intensities


# Calculate true vertical underground intensities


def _calc_u_intensities_tr(
    u_fluxes,
    depths,
    E_th_i,
    primary_model,
    interaction_model,
    atmosphere,
    location,
    month,
    output,
    file_name,
    force,
):

    """This function calculates true vertical underground intensities for flat overburdens."""

    # Check values

    constants._check_constants()

    if constants.get_overburden() != "flat":

        raise ValueError(
            "True vertical underground intensities can only be calculated for flat overburdens. See the Tutorial at {0} for an explanation.".format(
                "https://github.com/wjwoodley/mute/blob/main/docs/Tutorial.md#calculating-underground-intensities"
            )
        )

    if u_fluxes is None:

        raise Exception(
            "Underground intensities not calculated. The underground flux matrix was not calculated correctly."
        )

    # Calculate the underground intensities

    if constants.get_verbose() > 1:
        print("Calculating true vertical underground intensities.")

    u_fluxes_tr = u_fluxes[:, :, 0]
    u_intensities_tr = [
        scii.simpson(u_fluxes_tr[x, E_th_i:], constants.ENERGIES[E_th_i:])
        for x in range(len(constants._SLANT_DEPTHS))
    ]

    if constants.get_verbose() > 1:
        print("Finished calculating true vertical underground intensities.")

    # Interpolate to the slant depths given by the set vertical depth or the user-defined slant depths

    u_intensities_tr = np.exp(
        np.interp(depths, constants._SLANT_DEPTHS, np.log(u_intensities_tr))
    )

    # Write the results to the file

    if output:

        constants._check_directory(
            os.path.join(constants.get_directory(), "underground"), force=force
        )

        if file_name == "" or not isinstance(file_name, str):

            file_name = os.path.join(
                constants.get_directory(),
                "underground",
                "{0}_Underground_Intensities_tr.txt".format(constants.get_lab()),
            )

        file_out = open(file_name, "w")

        for x in range(len(depths)):

            file_out.write(
                "{0:1.2f} {1:1.14e}\n".format(depths[x], u_intensities_tr[x])
            )

        file_out.close()

        if constants.get_verbose() > 1:

            print(f"True vertical underground intensities written to {file_name}.")

    return u_intensities_tr


# Calculate double-differential underground intensities


def _calc_u_intensities_dd(
    u_fluxes,
    E_th_i,
    primary_model,
    interaction_model,
    atmosphere,
    location,
    month,
    output,
    file_name,
    force,
):

    """This function calculates double-differential underground intensities for mountain overburdens."""

    # Check values

    constants._check_constants()

    if constants.get_overburden() != "mountain":

        raise ValueError(
            "Double-differential underground intensities can only be calculated for non-flat overburdens. See the Tutorial at {0} for an explanation.".format(
                "https://github.com/wjwoodley/mute/blob/main/docs/Tutorial.md#calculating-underground-intensities"
            )
        )

    if u_fluxes is None:

        raise Exception(
            "Underground intensities not calculated. The underground flux matrix was not calculated correctly."
        )

    # Calculate the underground intensities

    u_intensities = np.zeros(
        (len(constants._SLANT_DEPTHS), len(constants.ANGLES_FOR_S_FLUXES))
    )

    if constants.get_verbose() > 1:
        print("Calculating underground intensities.")

    for x in range(len(constants._SLANT_DEPTHS)):

        for j in range(len(constants.ANGLES_FOR_S_FLUXES)):

            u_intensities[x, j] = scii.simpson(
                u_fluxes[x, E_th_i:, j], constants.ENERGIES[E_th_i:]
            )

    if constants.get_verbose() > 1:
        print("Finished calculating underground intensities.")

    # Create an interpolator object

    try:

        _interpolator

    except NameError:

        _interpolator = _create_interpolator(u_intensities)

    # Interpolate the intensities to the slant depths and zenith angles in the mountain profile file
    # Stack the output into an array for reshaping below

    interpolation_result = np.hstack(
        [
            np.exp(_interpolator(depth, angle))
            for depth, angle in zip(
                constants._mountain_slant_depths_all, constants._mountain_zenith_all
            )
        ]
    )

    # Calculate the intensities at the zenith and azimuthal angles specified in the mountain profile file

    interp_u_intensities = np.nan_to_num(
        np.reshape(
            interpolation_result,
            (len(constants.mountain.zenith), len(constants.mountain.azimuthal)),
        )
    )

    # Write the results to the file

    if output:

        constants._check_directory(
            os.path.join(constants.get_directory(), "underground"), force=force
        )

        if file_name == "" or not isinstance(file_name, str):

            file_name = os.path.join(
                constants.get_directory(),
                "underground",
                "{0}_Underground_Intensities_dd.txt".format(constants.get_lab()),
            )

        file_out = open(file_name, "w")

        for j in range(len(constants.mountain.zenith)):

            for az in range(len(constants.mountain.azimuthal)):

                file_out.write(
                    "{0:1.5f} {1:1.5f} {2:1.14e}\n".format(
                        constants.mountain.zenith[j],
                        constants.mountain.azimuthal[az],
                        interp_u_intensities[j, az],
                    )
                )

        file_out.close()

        if constants.get_verbose() > 1:
            print(f"Underground intensities written to {file_name}.")

    return interp_u_intensities


# Calculate total underground fluxes


def calc_u_tot_flux(force=False, **kwargs):

    """
    Calculate a total underground flux in units of [(cm^2 s)^-1].

    Parameters
    ----------
    force : bool
        If True, force the calculation of new matrices and the creation of new directories if required.

    Other Parameters
    -----------------
    u_fluxes : NumPy ndarray, optional (default: calcaulated from default surface fluxes and survival probabilities)
        An underground flux tensor of shape (28, 91, 20).

    s_fluxes : NumPy ndarray, optional (default: taken from surface.load_s_fluxes_from_file())
        A surface flux matrix of shape (91, 20).

    survival_probability_tensor : NumPy ndarray, optional (default: taken from propagation.load_survival_probability_tensor_from_file())
        A survival probability tensor of shape (91, 28, 91).

    angles : array-like, optional (default: taken from consants.angles)
        An array of zenith angles in [degrees] to calculate the underground intensities for. These are only used if method is "sd" or "eq".

    E_th : float, optional (default: 0)
        An energy threshold in [MeV].

    primary_model : str in {"GSF", "HG", "GH", "ZS", "ZSP", "PL27"} or tuple, optional (default: "GSF")
        The primary flux model to use in MCEq. Options:
        GSF  = GlobalSplineFitBeta
        HG   = HillasGaisser2012 (H3a)
        GH   = GaisserHonda
        ZS   = Zatsepin-Sokolskaya (Default)
        ZSP  = Zatsepin-Sokolskaya (PAMELA)
        PL27 = SimplePowerlaw27
        Alternatively, this can be set with a tuple. For example: (pm.GaisserStanevTilav, "3-gen").

    interaction_model : str, optional (default: "SIBYLL-2.3c")
        The hadronic interaction model to use in MCEq. See the Tutorial or MCEq documentation for a list of options.

    atmosphere : {"CORSIKA", "MSIS00"}, optional (default: "CORSIKA")
        The atmospheric model to use in MCEq. For US Standard Atmosphere, use "CORSIKA". For seasonal variations, use "MSIS00".

    location : str, optional (default: "USStd")
        The name of the location for which to calculate the surface fluxes. See the Tutorial or MCEq documentation for a list of options.

    month : str, optional (default: None)
        The month for which to calculate the surface fluxes. For US Standard Atmosphere, use None. For seasonal variations, use the month name.

    Returns
    -------
    u_tot_flux : float
        The total underground flux in units of [(cm^2 s)^-1].
    """

    # Check values

    constants._check_constants()

    # Calculate the total underground flux for a flat overburdens

    if constants.get_overburden() == "flat":

        # Get the angles keyword argument
        # For use in the integral over the angles

        angles = kwargs.get("angles", constants.angles)
        angles = np.atleast_1d(angles)

        # Calculate the underground intensities

        u_intensities = calc_u_intensities(method="sd", force=force, **kwargs)

        if u_intensities is None:

            raise Exception(
                "Total underground flux not calculated. The underground intensities were not calculated properly."
            )

        # Calculate the total underground flux
        # Because angles goes (0..89), cos(angles) goes (1..0)
        # cos(angles) is decreasing, but scii.simpson() wants an increasing array
        # Therefore, integrate backwards, using [::-1] on the integrand and steps
        # Otherwise, the answer will be negative

        u_tot_flux = (
            2
            * np.pi
            * scii.simpson(u_intensities[::-1], np.cos(np.radians(angles[::-1])))
        )

        return u_tot_flux

    # Calculate the total underground flux for mountains

    elif constants.get_overburden() == "mountain":

        # Calculate the underground intensities

        u_intensities = calc_u_intensities(method="dd", force=force, **kwargs)

        if u_intensities is None:

            raise Exception(
                "Total underground flux not calculated. The underground intensities were not calculated properly."
            )

        # Calculate the total underground flux
        # Integrate over all angles as defined by the mountain profile file

        u_tot_flux = scii.simpson(
            [
                scii.simpson(
                    u_intensities[j, :], np.radians(constants.mountain.azimuthal)
                )
                for j in range(len(constants.mountain.zenith))
            ][::-1],
            np.cos(np.radians(constants.mountain.zenith[::-1])),
        )

        return u_tot_flux

    else:

        raise NotImplementedError(
            'Overburdens of type {0} are not available. The only options are "flat" and "mountain".'.format(
                constants.get_overburden()
            )
        )



#########################
#########################
###                   ###
###  MUTE             ###
###  William Woodley  ###
###  15 July 2022     ###
###                   ###
#########################
#########################

# Import packages

import os

import numpy as np
from tqdm import tqdm

import mute.constants as constants

try:

    import proposal as pp

except ImportError:

    pass

# Create the propagator


def _create_propagator(force):

    """This function creates the propagator object in PROPOSAL for use in propagation._propagation_loop()"""

    # Check values

    constants._check_constants(force=force)

    # The creation of the propagator can be very slow the first time
    # The propagator is used in every iteration of the doubly-nested propagation loop
    # Make it a global variable so it only has to be created once

    global propagator

    if constants.get_verbose() > 1:
        print("Creating propagator.")

    # Propagator arguments

    mu = pp.particle.MuMinusDef()
    cuts = pp.EnergyCutSettings(500, 0.05, True)

    if constants.get_medium() == "rock":

        medium = pp.medium.StandardRock()

    elif constants.get_medium() == "water":

        medium = pp.medium.Water()

    elif constants.get_medium() == "ice":

        medium = pp.medium.Ice()

    elif constants.get_medium() == "air":

        medium = pp.medium.Air()

    else:

        raise NotImplementedError(
            "Medium type {0} not implemented. The only options are {1}, {2}, {3}, and {4}.".format(
                constants.get_medium(), "rock", "water", "ice", "air"
            )
        )

    args = {"particle_def": mu, "target": medium, "interpolate": True, "cuts": cuts}

    # Initialise standard cross-sections, then specify and set parametrisation models

    cross_sections = pp.crosssection.make_std_crosssection(**args)

    brems_param = pp.parametrization.bremsstrahlung.KelnerKokoulinPetrukhin(lpm=False)
    epair_param = pp.parametrization.pairproduction.KelnerKokoulinPetrukhin(lpm=False)
    ionis_param = pp.parametrization.ionization.BetheBlochRossi(energy_cuts=cuts)
    shado_param = pp.parametrization.photonuclear.ShadowButkevichMikheyev()
    photo_param = pp.parametrization.photonuclear.AbramowiczLevinLevyMaor97(
        shadow_effect=shado_param
    )

    cross_sections[0] = pp.crosssection.make_crosssection(brems_param, **args)
    cross_sections[1] = pp.crosssection.make_crosssection(epair_param, **args)
    cross_sections[2] = pp.crosssection.make_crosssection(ionis_param, **args)
    cross_sections[3] = pp.crosssection.make_crosssection(photo_param, **args)

    # Propagation utility

    collection = pp.PropagationUtilityCollection()

    collection.interaction = pp.make_interaction(cross_sections, True)
    collection.displacement = pp.make_displacement(cross_sections, True)
    collection.time = pp.make_time(cross_sections, mu, True)
    collection.decay = pp.make_decay(cross_sections, mu, True)

    pp.PropagationUtilityCollection.cont_rand = False

    utility = pp.PropagationUtility(collection=collection)

    # Other settings

    pp.do_exact_time = False

    # Set up geometry

    detector = pp.geometry.Sphere(
        position=pp.Cartesian3D(0, 0, 0), radius=10000000, inner_radius=0
    )
    density_distr = pp.density_distribution.density_homogeneous(
        mass_density=constants.get_density()
    )

    propagator = pp.Propagator(mu, [(detector, utility, density_distr)])

    if constants.get_verbose() > 1:
        print("Finished creating propagator.")

    return propagator


# Propagation function


def _propagation_loop(energy, slant_depth, force):

    """This function propagates n_muon muons, looping over the energies and slant depths, and returns the muons' underground energies in [MeV]."""

    # Check values

    constants._check_constants(force=force)

    n_muon = constants.get_n_muon()

    # Convert the slant depth from [km.w.e.] to [cm]

    convert_to_cm = 1e5 * 0.997 / constants.get_density()

    # Initialise a list of underground energies

    u_energies_ix = []

    # Define the initial state of the muon

    mu_initial = pp.particle.ParticleState()
    mu_initial.energy = energy + constants._MU_MASS
    mu_initial.position = pp.Cartesian3D(0, 0, 0)
    mu_initial.direction = pp.Cartesian3D(0, 0, -1)

    # Propagate n_muon muons

    for _ in range(n_muon):

        # Propagate the muons

        track = propagator.propagate(mu_initial, slant_depth * convert_to_cm)

        # Test whether or not the muon has energy left (has not lost all of its energy or has not decayed)
        # If it does, record its energy
        # If it does not, ignore this muon and proceed with the next loop iteration

        if (
            track.track_energies()[-1] != constants._MU_MASS
            and track.track_types()[-1] != pp.particle.Interaction_Type.decay
        ):

            # Store the final underground energy of the muon

            u_energies_ix.append(track.track_energies()[-1])

    # Return the underground energies for the muon

    return u_energies_ix


# Propagate the muons and return underground energies


def propagate_muons(seed=0, job_array_number=0, output=None, force=False):

    """
    Propagate muons for the default surface energy grid and slant depths.

    The default surface energy grid is given by constants.ENERGIES, and the default slant depths are given by constants._SLANT_DEPTHS.

    Parameters
    ----------
    seed : int, optional (default: 0)
        The random seed for use in the PROPOSAL propagator.

    job_array_number : int, optional (default: 0)
        The job array number from a high-statistics run on a computer cluster. This is set so the underground energy files from each job in the job array will be named differently.

    output : bool, optional (default: taken from constants.get_output())
        If True, an output file will be created to store the results.

    force : bool, optional (default: False)
        If True, force the creation of new directories if required.

    Returns
    -------
    u_energies : NumPy ndarray
        A two-dimensional array containing lists of underground energies for muons that survived the propagation. The shape of the array will be (91, 28), and the underground energies will be in units of [MeV].
    """

    # Check values

    constants._check_constants(force=force)

    if output is None:
        output = constants.get_output()

    assert type(job_array_number) == int, "job_array_number must be an integer."

    # Create the propagator once

    _create_propagator(force=force)

    # Set the random seed

    pp.RandomGenerator.get().set_seed(seed)

    # Initialise a matrix of underground energies

    u_energies = np.zeros(
        (len(constants.ENERGIES), len(constants._SLANT_DEPTHS)), dtype=np.ndarray
    )

    # Run the propagation function

    if constants.get_verbose() >= 1:

        print(
            "Propagating {0} muons.".format(
                constants.get_n_muon()
                * len(constants.ENERGIES)
                * len(constants._SLANT_DEPTHS)
            )
        )

    for i in (
        tqdm(range(len(constants.ENERGIES)))
        if constants.get_verbose() >= 1
        else range(len(constants.ENERGIES))
    ):

        for x in range(len(constants._SLANT_DEPTHS)):

            u_energies[i, x] = _propagation_loop(
                constants.ENERGIES[i], constants._SLANT_DEPTHS[x], force=force
            )

    if constants.get_verbose() >= 1:
        print("Finished propagation.")

    # Write the results to the file

    if output:

        constants._check_directory(
            os.path.join(constants.get_directory(), "underground_energies"), force=force
        )

        file_name = os.path.join(
            constants.get_directory(),
            "underground_energies",
            "{0}_{1}_{2}_Underground_Energies_{3}.npy".format(
                constants.get_medium(),
                constants.get_density(),
                constants.get_n_muon(),
                job_array_number,
            ),
        )

        np.save(file_name, u_energies)

        if constants.get_verbose() > 1:
            print(f"Underground energies written to {file_name}.")

    return u_energies


# Load underground energies


def _load_u_energies_from_files(file_name_pattern, n_job=1, force=False):

    """
    Load the underground energies resulting from the PROPOSAL Monte Carlo from a file or collection of files stored in data/underground_energies.

    Parameters
    ----------
    file_name_pattern : str
        The file name pattern for the file(s) that the underground energy data is stored in. This should end before an underscore so the function can append the job array number. For example, pass "underground_energies" for a set of files beginning with "underground_energies_0.npy".

    n_job : int, optional (default: 1)
        The number of jobs that were run on the computer cluster. Set this to the number of files the underground energies are spread across.

    force : bool, optional (default: False)
        If True, force the propagation of muons and the creation of new directories if required.

    Returns
    -------
    u_energies : NumPy ndarray
        A two-dimensional array containing lists of underground energies for muons that survived the propagation. The shape of the array will be (91, 28), and the underground energies will be in units of [MeV].
    """

    # Check values

    constants._check_constants(force=force)

    # Check that the directory exists

    if not os.path.exists(
        os.path.join(constants.get_directory(), "underground_energies")
    ):

        if constants.get_verbose() >= 1:

            print(
                f"{constants.get_directory}/underground_energies does not exist. Underground energies not loaded."
            )

        return

    # Test if the file exists

    if not os.path.isfile(f"{file_name_pattern}_0.npy"):

        if constants.get_verbose() >= 1:

            print(
                f"{file_name_pattern}_0.npy does not exist. Underground energies not loaded."
            )

        return

    # Fill a u_energies array with empty lists that will be able to be extended

    u_energies = np.empty(
        (len(constants.ENERGIES), len(constants._SLANT_DEPTHS)), dtype=object
    )

    for i in np.ndindex(u_energies.shape):
        u_energies[i] = []

    # Loop over all output files and add the contents to u_energies

    for job_array_number in (
        tqdm(range(n_job)) if constants.get_verbose() >= 1 else range(n_job)
    ):

        u_energies += np.load(
            f"{file_name_pattern}_{job_array_number}.npy", allow_pickle=True
        )

    if constants.get_verbose() > 1:
        print("Loaded underground energies.")

    return u_energies


# Calculate survival probabilities


def calc_survival_probability_tensor(
    seed=0, file_name_pattern=None, n_job=1, output=None, file_name="", force=False
):

    """
    Calculate survival probabilities in units of [(MeV^2 km.w.e.)^-1] for the default surface energy grid and slant depths.

    The default surface energy grid is given by constants.ENERGIES, and the default slant depths are given by constants._SLANT_DEPTHS. If the propagation of muons has already been done, this will load the underground energies file, unless force is set to True.

    Parameters
    ----------
    seed : int, optional (default: 0)
        The random seed for use in the PROPOSAL propagator.

    file_name_pattern : str, optional
        The file name pattern for the file(s) that the underground energy data is stored in. This should end before an underscore so the function can append the job array number. For example, pass "underground_energies" for a set of files beginning with "underground_energies_0.npy".

    n_job : int, optional (default: 1)
        The number of jobs that were run on the computer cluster. Set this to the number of files the underground energy data is spread across.

    output : bool, optional (default: taken from constants.get_output())
        If True, an output file will be created to store the results.

    file_name : str, optional (default: constructed from set global propagation constants)
        The name of the file in which to store the results. If output is False or None, this is ignored.

    force : bool, optional (default: False)
        If True, this will force the muons to be propagated in the Monte Carlo whether an underground energies file already exists or not.

    Returns
    -------
    survival_probability_tensor : NumPy ndarray
        A three-dimensional array containing the survival probabilities. The shape of the array will be (91, 28, 91), and the survival probabilities will be in units of [(MeV^2 km.w.e.)^-1].
    """

    # Check values

    if output is None:
        output = constants.get_output()

    # Construct a default file name pattern

    file_name_pattern_default = os.path.join(
        constants.get_directory(),
        "underground_energies",
        "{0}_{1}_{2}_Underground_Energies".format(
            constants.get_medium(),
            constants.get_density(),
            int(constants.get_n_muon() / n_job),
        ),
    )

    # Check if propagate_muons() should be forced or not
    # If not, check if underground energy files exist
    # If not, ask if muons should be propagated

    if force:

        u_energies = propagate_muons(seed=seed, output=output, force=force)

    else:

        # Check if the user has specified underground energies to load
        # If not, look for the default file name pattern and check if it exists
        # If so, load the underground energies
        # If not, ask if muons should be propagated

        if file_name_pattern is not None:

            u_energies = _load_u_energies_from_files(
                file_name_pattern=os.path.join(
                    constants.get_directory(), "underground_energies", file_name_pattern
                ),
                n_job=n_job,
                force=force,
            )

        elif os.path.isfile(f"{file_name_pattern_default}_0.npy"):

            u_energies = _load_u_energies_from_files(
                file_name_pattern=file_name_pattern_default, n_job=n_job, force=force
            )

        else:

            answer = input(
                "No underground energy file currently exists for the set medium, density, or number of muons. Would you like to create one (y/n)?: "
            )

            if answer.lower() == "y":

                u_energies = propagate_muons(seed=seed, output=output, force=force)

            else:

                print("Underground energies not calculated.")
                print("Survival probabilities not calculated.")

                return

    # Check that the underground energies were loaded

    if u_energies is None:

        raise Exception(
            "Survival probabilities not calculated. The underground energies were not loaded or calculated correctly."
        )

    # Calculate the survival probabilities
    # Zeroth index = Surface energy
    # First index  = Slant depth
    # Second index = Underground energy

    survival_probability_tensor = np.zeros(
        (len(constants.ENERGIES), len(constants._SLANT_DEPTHS), len(constants.ENERGIES))
    )

    if constants.get_verbose() > 1:
        print("Calculating survival probabilities.")

    # Loop over the surface energies and slant depths
    # Histogram the underground energies to count how many are in each underground energy bin
    # Divide the counts by the number of muons
    # This counts how many muons survived per bin out of the total number that was thrown

    for i in range(len(constants.ENERGIES)):

        for x in range(len(constants._SLANT_DEPTHS)):

            survival_probability_tensor[i, x, :] = np.histogram(
                np.array(u_energies[i, x]), bins=constants._E_BINS
            )[0] / float(constants.get_n_muon())

    if constants.get_verbose() > 1:
        print("Finished calculating survival probabilities.")

    # Write the results to the file

    if output:

        constants._check_directory(
            os.path.join(constants.get_directory(), "survival_probabilities"),
            force=force,
        )

        if file_name == "" or not isinstance(file_name, str):

            file_name = os.path.join(
                constants.get_directory(),
                "survival_probabilities",
                "{0}_{1}_{2}_Survival_Probabilities.txt".format(
                    constants.get_medium(),
                    constants.get_density(),
                    constants.get_n_muon(),
                ),
            )

        file_out = open(file_name, "w")

        for i in range(len(constants.ENERGIES)):

            for x in range(len(constants._SLANT_DEPTHS)):

                for u in range(len(constants.ENERGIES)):

                    file_out.write(
                        "{0:1.14f} {1:1.5f} {2:1.14f} {3:1.14e}\n".format(
                            constants.ENERGIES[i],
                            constants._SLANT_DEPTHS[x],
                            constants.ENERGIES[u],
                            survival_probability_tensor[i, x, u],
                        )
                    )

        file_out.close()

        if constants.get_verbose() > 1:
            print(f"Survival probabilities written to {file_name}.")

    return survival_probability_tensor


def load_survival_probability_tensor_from_file(file_name="", force=False):

    """
    Retrieve a survival probability tensor in units of [(MeV^2 km.w.e.)^-1] stored in data/survival_probabilities based on the set global propagation constants.

    The function searches for a file name that matches the set medium, density, and number of muons. If the file does not exist, prompt the user to run propagation.calc_survival_probability_tensor().

    Parameters
    ----------
    file_name : str, optional (default: constructed from set global propagation constants)
        The name of the file in which to store the results. If output is False or None, this is ignored.

    force : bool
        If True, force the calculation of a new survival probability tensor and the creation of new directories if required.

    Returns
    -------
    survival_probability_tensor : NumPy ndarray
        A three-dimensional array containing the survival probabilities. The shape of the array will be (91, 28, 91), and the survival probabilities will be in units of [(MeV^2 km.w.e.)^-1].
    """

    # Check values

    constants._check_constants(force=force)

    # Check if a survival probability tensor has already been loaded
    # If so, return the tensor that has already been loaded
    # If not, load a tensor based on the set global propagation constants

    if (
        constants._current_survival_probability_tensor is not None
        and constants._survival_probability_tensor_configuration
        == {
            "medium": constants.get_medium(),
            "density": constants.get_density(),
            "n_muon": constants.get_n_muon(),
        }
    ):

        if constants.get_verbose() > 1:
            print(
                "Survival probabilities already loaded for {0} with density {1} gcm^-3 and {2} muons.".format(
                    constants._survival_probability_tensor_configuration["medium"],
                    constants._survival_probability_tensor_configuration["density"],
                    constants._survival_probability_tensor_configuration["n_muon"],
                )
            )

        return constants._current_survival_probability_tensor

    # Define a function to update the survival probability tensor cache

    def update_cache(survival_probability_tensor):

        constants._survival_probability_tensor_configuration = {
            "medium": constants.get_medium(),
            "density": constants.get_density(),
            "n_muon": constants.get_n_muon(),
        }
        constants._current_survival_probability_tensor = survival_probability_tensor

    # Define a function to run if there is no survival probability file

    def no_file(force):

        # If the file does not exist, ask the user if they want to run PROPOSAL to create it

        if not force:

            answer = input(
                "No survival probability matrix currently exists for the set medium, density, or number of muons. Would you like to create one (y/n)?: "
            )

        if force or answer.lower() == "y":

            survival_probability_tensor = calc_survival_probability_tensor(force=force)

            update_cache(survival_probability_tensor)

            return survival_probability_tensor

        else:

            print("Survival probabilities not calculated.")

            return

    # Construct a file name based on the set medium, density, and number of muons if one has not been provided

    if file_name == "" or not isinstance(file_name, str):

        file_name = os.path.join(
            constants.get_directory(),
            "survival_probabilities",
            "{0}_{1}_{2}_Survival_Probabilities.txt".format(
                constants.get_medium(), constants.get_density(), constants.get_n_muon()
            ),
        )

    # Check if the file exists

    if os.path.isfile(file_name):

        if constants.get_verbose() > 1:
            print(f"Loading survival probabilities from {file_name}.")

        # Check that the file has the correct numbers of energies and slant depths

        file = open(file_name, "r")
        n_lines = len(file.read().splitlines())
        file.close()

        # If so, read in the survival probabilities in from it

        if n_lines == len(constants.ENERGIES) * len(constants._SLANT_DEPTHS) * len(
            constants.ENERGIES
        ):

            survival_probability_tensor = np.reshape(
                np.loadtxt(file_name)[:, 3],
                (
                    len(constants.ENERGIES),
                    len(constants._SLANT_DEPTHS),
                    len(constants.ENERGIES),
                ),
            )

            if constants.get_verbose() > 1:
                print("Loaded survival probabilities.")

            update_cache(survival_probability_tensor)

            return survival_probability_tensor

        # If the file does not have the correct numbers of energies and slant depths, run the no_file() function

        else:

            return no_file(force=force)

    # If the file does not exist, run the no_file() function

    else:

        return no_file(force=force)



import os

import numpy as np

import mute.constants as mtc
import mute.surface as mts
import mute.underground as mtu


def test_s_tot_flux():

    mtc.clear()

    s_fluxes = mts.load_s_fluxes_from_file(primary_model="HG")

    s_tot_flux_calc = mts.calc_s_tot_flux(s_fluxes=s_fluxes)
    s_tot_flux_read = 0.011555347252101776

    assert np.allclose(s_tot_flux_calc, s_tot_flux_read)


def test_u_tot_flux_flat():

    mtc.clear()

    mtc.set_output(False)
    mtc.set_overburden("flat")
    mtc.set_vertical_depth(3.5)

    u_fluxes = mtu.calc_u_fluxes(full_tensor=True)

    u_tot_flux_calc = mtu.calc_u_tot_flux(u_fluxes=u_fluxes, E_th=100, force=True)
    u_tot_flux_read = 1.123374482778583e-08

    assert np.allclose(u_tot_flux_calc, u_tot_flux_read)


def test_u_tot_flux_mountain():

    mtc.clear()

    mtc.set_output(False)
    mtc.set_overburden("mountain")

    mtc.load_mountain("example_mountain_profile.txt")

    u_tot_flux_calc = mtu.calc_u_tot_flux(interaction_model="DDM", force=True)
    u_tot_flux_read = 1.4121546671702574e-06

    assert np.allclose(u_tot_flux_calc, u_tot_flux_read)



import os
from shutil import copytree, rmtree

GitHub_data_file = "data_20220715"


def download_file(url, dir_out):

    """Download the data files from GitHub."""

    import requests
    from zipfile import ZipFile

    # Download the zip file

    response = requests.get(url, stream=True)
    file_out = "mute_data_files.zip"

    with open(file_out, "wb") as f:

        for line in response.iter_content(chunk_size=1024 * 1024):

            f.write(line)

    # Unzip the file

    with ZipFile(file_out, "r") as f:

        f.extractall(dir_out)

    # Rename the data directory "data"
    # If it already exists, move the unzipped files into it and replace existing ones with the same name
    # If it does not, rename the unzipped directory "data"

    os.chdir(dir_out)

    if os.path.isdir("data"):

        copytree(GitHub_data_file, "data", dirs_exist_ok=True)
        rmtree(GitHub_data_file)

    else:

        os.rename(GitHub_data_file, "data")

    os.chdir("..")

    # Delete the zip file

    if os.path.isfile(file_out):
        os.remove(file_out)


# Download the files from GitHub to the directory MUTE is being run from

if not os.path.isfile(
    os.path.join(os.path.dirname(__file__), "data", f"{GitHub_data_file}.txt")
):

    download_file(
        f"https://github.com/wjwoodley/mute/releases/download/0.1.0/{GitHub_data_file}.zip",
        os.path.dirname(__file__),
    )



#########################
#########################
###                   ###
###  MUTE             ###
###  William Woodley  ###
###  15 July 2022     ###
###                   ###
#########################
#########################

# Import packages

import os
from collections import namedtuple

import numpy as np

# Energies in [MeV]

_E_BINS = np.logspace(1.9, 14, 122)[:-30]  # Bin edges
_E_WIDTHS = np.diff(_E_BINS)  # Bin widths
ENERGIES = np.sqrt(_E_BINS[1:] * _E_BINS[:-1])  # Bin centers

# Slant depths in [km.w.e.] and angles in [degrees]
# These are the defaults used to construct the matrices
# The user can enter their own angles in the function calls, which will interpolate over the grids created by these angles

_X_MIN = 0.5
_X_MAX = 14

_SLANT_DEPTHS = np.linspace(_X_MIN, _X_MAX, int(2 * (_X_MAX - _X_MIN) + 1))
_ANGLES = np.degrees(np.arccos(_X_MIN / _SLANT_DEPTHS))

slant_depths = _SLANT_DEPTHS
angles = _ANGLES

# Angles in [degrees] specifically for the calculation of surface flux matrices

ANGLES_FOR_S_FLUXES = np.linspace(0, 90, 20)

# Other constants
# The rest mass of a muon in [MeV]
# Months of the year

_MU_MASS = 105.6583745

MONTHS = [
    "January",
    "February",
    "March",
    "April",
    "May",
    "June",
    "July",
    "August",
    "September",
    "October",
    "November",
    "December",
]
MONTHS_SNAMES = [
    "Jan.",
    "Feb.",
    "Mar.",
    "Apr.",
    "May.",
    "Jun.",
    "Jul.",
    "Aug.",
    "Sep.",
    "Oct.",
    "Nov.",
    "Dec.",
]

# Define default user-set global constants

_verbose = 2
_output = True
_directory = os.path.join(os.path.dirname(__file__), "data")
_lab = "Default"
_overburden = "flat"
_vertical_depth = _X_MIN
_medium = "rock"
_density = 2.65
_n_muon = 100000

# Keep track of which survival probability tensor is loaded

_survival_probability_tensor_configuration = {}
_current_survival_probability_tensor = None

# Keep track of whether or not a mountain profile has been loaded

_mountain_loaded = False

# Use extrapolation for depths lower than 0.5 km.w.e.
# This limitation comes from PROPOSAL, not MUTE

shallow_extrapolation = False

# Setters and getters for global constants

# Verbose


def set_verbose(verbose):

    """
    Set the verbosity level.

    Parameters
    ----------
    verbose : int (default: 2)
        The verbosity level. Options:
        0 = Print no information.
        1 = Print the progress through the calculation of surface fluxes if calculating surface fluxes, the propagation loop if propagating muons, or the loading of the Monte Carlo underground energies from files if loading underground energies.
        2 = Print when calculations start and finish, and where data is read from and written to.
    """

    assert isinstance(verbose, int)

    global _verbose

    _verbose = verbose


def get_verbose():

    """Return the set verbosity level."""

    return _verbose


# Whether or not the results are written to output files in directory


def set_output(output):

    """
    Set whether results are output to files or not. The default is True.

    Most functions also have their own optional output keyword argument to give more control over which output files are created.
    """

    assert isinstance(output, bool)

    global _output

    _output = output


def get_output():

    """Return the set output setting."""

    return _output


# Directory for input and output


def set_directory(directory):

    """Set the working directory for files to be written to and read from. The default is \"data\"."""

    from .__init__ import download_file

    global _directory

    _directory = directory

    if not os.path.isfile(os.path.join(_directory, "data_20220715.txt")):

        download_file(
            "https://github.com/wjwoodley/mute/releases/download/0.1.0/data_20220715.zip",
            _directory,
        )


def get_directory():

    """Return the set working directory for files to be written to and read from."""

    return _directory


# Lab name used in output files


def set_lab(lab):

    """Set the name of the lab. This is used in the output file names in the \"underground\" data directory as a way to differentiate different data files. The default is \"Default\"."""

    global _lab

    _lab = lab


def get_lab():

    """Return the set name of the lab."""

    return _lab


# Overburden


def set_overburden(overburden):

    """Set the overburden type. The default is  \"flat\" overburden."""

    assert overburden in [
        "flat",
        "mountain",
    ], 'overburden must be set to either "flat" or "mountain".'

    global _overburden

    _overburden = overburden


def get_overburden():

    """Return the set overburden type."""

    return _overburden


# Vertical depth in [km.w.e.]


def set_vertical_depth(vertical_depth):

    """Set the vertical depth, h. The default is 0.5 km.w.e."""

    global _vertical_depth
    global slant_depths
    global angles

    _vertical_depth = vertical_depth

    # Use the set vertical depth to calculate new slant depths and zenith angles
    # Only do this for flat overburdens
    # For mountains, the slant depths and angles will be calculated in load_mountain()

    if vertical_depth < _SLANT_DEPTHS[0] and not shallow_extrapolation:

        raise Exception(
            "The minimum default available slant depth is 0.5 km.w.e. Set constants.shallow_extrapolation to True to enable calculations for depths lower than 0.5 km.w.e. (not recommended)."
        )

    if get_overburden() == "flat":

        slant_depths = np.sort(
            np.concatenate(
                ([_vertical_depth], _SLANT_DEPTHS[_SLANT_DEPTHS > _vertical_depth])
            )
        )
        angles = np.degrees(np.arccos(_vertical_depth / slant_depths))


def get_vertical_depth():

    """Return the set vertical depth."""

    return _vertical_depth


# Medium


def set_medium(medium):

    """Set the medium for the muons to be propagated through. The default is standard rock (\"rock\")."""

    assert medium in [
        "rock",
        "water",
        "ice",
        "air",
    ], 'medium must be set to "rock", "water", "ice", or "air".'

    global _medium

    _medium = medium


def get_medium():

    """Return the set propagation medium."""

    return _medium


# Density in [gcm^-3]


def set_density(density):

    """Set the density of the propagation medium. The default is 2.65 gcm^-3 (the density of standard rock)."""

    assert density > 0, "Media density must be positive."

    global _density

    _density = density


def get_density():

    """Return the set density of the propagation medium."""

    return _density


# Number of muon


def set_n_muon(n_muon):

    """Set the number of muons to be propagated per surface energy-slant depth bin. The default is 100000, as the provided default survival probability tensors in the data directory are given for 100000 muons."""

    assert isinstance(n_muon, int), "Number of muons must be an integer."
    assert n_muon > 0, "Number of muons must be positive."

    global _n_muon

    _n_muon = n_muon


def get_n_muon():

    """Return the set number of muons to be propagated per surface energy-slant depth bin."""

    return _n_muon


# Check if required directories exist


def _check_directory(directory, force=False):

    """This function checks whether a directory required to store output files exists or not. If it does not, ask the user if it should be created."""

    if not os.path.exists(directory):

        # If the directory does not exist, ask the user if they want to create it

        if not force:

            answer = input(
                f"{directory} does not currently exist. Would you like to create it (y/n)?: "
            )

        if force or answer.lower() == "y":

            os.makedirs(directory)

        else:

            print("Directory not created.")

            return

    else:

        return


# Check that the constants are set to correctly before running any calculation functions


def _check_constants(force=False):

    """This function checks that the constants are set correctly before running any calculation functions."""

    # Check that the working directory the user has set exists

    _check_directory(get_directory(), force=force)

    # Check that the overburden has been set to one of the available options

    if _overburden is None:

        raise ValueError('Overburden must be set to either "flat" or "mountain".')

    # Check that the vertical depth is set correctly, if needed

    elif get_overburden() == "flat":

        assert get_vertical_depth() is not None, "Initial vertical depth must be set."
        assert get_vertical_depth() > 0, "Initial vertical depth must be positive."

        assert len(slant_depths) == len(angles) and np.allclose(
            slant_depths, get_vertical_depth() / np.cos(np.radians(angles))
        ), "slant_depths must correspond to angles. Do not assign constants.slant_depths or constants.angles directly. Instead, use constants.set_vertical_depth() in combination with the angles and depths parameters in individual functions. Run constants.clear() to reset the values of slant_depths and / or angles."

    # Check that a mountain profile has been loaded, if needed

    elif get_overburden() == "mountain":

        if not _mountain_loaded:

            raise Exception(
                "The mountain profile must be loaded first by passing a txt file to constants.load_mountain()."
            )

    else:

        raise NotImplementedError(
            'Overburdens of type {0} are not available. The only options are "flat" and "mountain".'.format(
                constants.get_overburden()
            )
        )

    # Check that the number of muons is set correctly

    if _n_muon <= 0:

        raise ValueError("Number of muons must be a positive integer.")


# Load mountain data


def load_mountain(file_name, units="kmwe", density=get_density(), max_slant_depth=14):

    """
    Load mountain data from a profile file.

    The first column should be the zenith angle in [degrees]; the second column should be the azimuthal angle in [degrees]; the third column should be the slant depth in units of units (see below). This function makes available the variables listed under "Sets" below.

    Parameters
    ----------
    file_name : str
        The full path and name of the .txt file containing the profile information of the mountain.

    units : str, optional (default: "kmwe")
        The units of the slant depths in file_name. This must be one of ["m", "km", "mwe", "kmwe"].

    density : float, optional (default: taken from constants.get_density())
        The density in [g cm^-3] of the rock for which the depths are specified in the file.

    max_slant_depth : float, optional (default: 14)
        The maximum slant depth in [km.w.e.] to take from the file. Any data for slant depths above this value will be set to 0 km.w.e. and will be ignored throughout calculations. The default is 14 km.w.e., consistent with the maximum slant depth in constants._SLANT_DEPTHS being 14 km.w.e.

    Sets
    ----
    constants.mountain.zenith : NumPy ndarray
        Unique zenith angles in [degrees] sorted from smallest to largest.

    constants.mountain.azimuthal : NumPy ndarray
        Unique azimuthal angles in [degrees] sorted from smallest to largest.

    constants.mountain.slant_depths : NumPy ndarray
        Unique slant depths in [km.w.e.] in a matrix of shape (len(constants.mountain_zenith), len(constants.mountain_azimuthal)).
    """

    # Check values

    assert (
        get_overburden() == "mountain"
    ), 'The overburden type must be set to "mountain".'
    assert units in [
        "m",
        "km",
        "mwe",
        "kmwe",
    ], 'Units must be "m", "km", "mwe", or "kmwe".'

    # Global variables

    global _mountain_loaded
    global _mountain_zenith_all
    global _mountain_azimuthal_all
    global _mountain_slant_depths_all

    global mountain

    # Create a named tuple to hold immutable results

    Mountain = namedtuple("Mountain", ("zenith", "azimuthal", "slant_depths"))

    # If the data is in [m] or [km], convert to [m.w.e.] or [km.w.e.]

    density_mult = 1

    if units == "m" or units == "km":
        density_mult = density / 0.997

    # If the data is in [m] or [m.w.e.], convert to [km.w.e.]

    scale = 1

    if units == "m" or units == "mwe":
        scale = 1e-3

    # Load the mountain profile data from the file

    _mountain_zenith_all = np.loadtxt(file_name)[:, 0]
    _mountain_azimuthal_all = np.loadtxt(file_name)[:, 1]
    _mountain_slant_depths_all = np.loadtxt(file_name)[:, 2] * density_mult * scale

    # Extract the unique angles and slant depths

    mountain_zenith = np.unique(_mountain_zenith_all)
    mountain_azimuthal = np.unique(_mountain_azimuthal_all)
    mountain_slant_depths = np.reshape(
        np.nan_to_num(_mountain_slant_depths_all),
        (len(mountain_zenith), len(mountain_azimuthal)),
    )

    # Remove the slant depths above the maximum slant depth threshold by setting them to 0

    mountain_slant_depths[mountain_slant_depths > max_slant_depth] = 0

    # Assign the elements of the namedtuple

    mountain = Mountain(mountain_zenith, mountain_azimuthal, mountain_slant_depths)

    # If the mountain profile loads successfully, set _mountain_loaded to True

    _mountain_loaded = True


# Clear all variables


def clear():

    """Reset all of the values set or calculated in MUTE to their default values."""

    # Import packages

    import gc

    # Energies, slant depths, and zenith angles

    global _E_BINS
    global _E_WIDTHS
    global ENERGIES
    global _X_MIN
    global _X_MAX
    global _SLANT_DEPTHS
    global _ANGLES
    global slant_depths
    global angles
    global ANGLES_FOR_S_FLUXES

    _E_BINS = np.logspace(1.9, 14, 122)[:-30]  # [MeV]
    _E_WIDTHS = np.diff(_E_BINS)  # [MeV]
    ENERGIES = np.sqrt(_E_BINS[1:] * _E_BINS[:-1])  # [MeV]
    _X_MIN = 0.5  # [km.w.e.]
    _X_MAX = 14  # [km.w.e.]
    _SLANT_DEPTHS = np.linspace(
        _X_MIN, _X_MAX, int(2 * (_X_MAX - _X_MIN) + 1)
    )  # [km.w.e.]
    _ANGLES = np.degrees(np.arccos(0.5 / _SLANT_DEPTHS))  # [degrees]
    slant_depths = _SLANT_DEPTHS  # [km.w.e.]
    angles = _ANGLES  # [degrees]
    ANGLES_FOR_S_FLUXES = np.linspace(0, 90, 20)  # [degrees]

    # Other constants and variables

    global _MU_MASS
    global MONTHS
    global MONTHS_SNAMES
    global _survival_probability_tensor_configuration
    global _current_survival_probability_tensor
    global _mountain_loaded
    global shallow_extrapolation

    _MU_MASS = 105.6583745  # [MeV]
    MONTHS = [
        "January",
        "February",
        "March",
        "April",
        "May",
        "June",
        "July",
        "August",
        "September",
        "October",
        "November",
        "December",
    ]
    MONTHS_SNAMES = [
        "Jan",
        "Feb",
        "Mar",
        "Apr",
        "May",
        "Jun",
        "Jul",
        "Aug",
        "Sep",
        "Oct",
        "Nov",
        "Dec",
    ]
    _survival_probability_tensor_configuration = {}
    _current_survival_probability_tensor = None
    _mountain_loaded = False
    shallow_extrapolation = False

    # Global constants and variables

    global _verbose
    global _output
    global _directory
    global _lab
    global _overburden
    global _vertical_depth
    global _medium
    global _density
    global _n_muon

    _verbose = 2
    _output = True
    _directory = os.path.join(os.path.dirname(__file__), "data")
    _lab = "Default"
    _overburden = "flat"
    _vertical_depth = _X_MIN
    _medium = "rock"
    _density = 2.65
    _n_muon = 100000

    # Global mountain variables
    # Set the variables to None first, in case they do not already exist in the namespace

    global _mountain_zenith_all
    global _mountain_azimuthal_all
    global _mountain_slant_depths_all
    global mountain

    _mountain_zenith_all = None
    _mountain_azimuthal_all = None
    _mountain_slant_depths_all = None
    mountain = None

    del _mountain_zenith_all
    del _mountain_azimuthal_all
    del _mountain_slant_depths_all
    del mountain

    gc.collect()



#########################
#########################
###                   ###
###  MUTE             ###
###  William Woodley  ###
###  15 July 2022     ###
###                   ###
#########################
#########################

# Import packages

import os

import numpy as np
import scipy.integrate as scii
from tqdm import tqdm

import mute.constants as constants

# Calculate surface fluxes


def calc_s_fluxes(
    primary_model="GSF",
    interaction_model="SIBYLL-2.3c",
    atmosphere="CORSIKA",
    location="USStd",
    month=None,
    output=None,
    file_name="",
    force=False,
    test=False,
):

    """
    Calculate surface fluxes in units of [(cm^2 s sr MeV)^-1] for default surface energy grid and surface flux zenith angles.

    The default surface energy grid is given by constants.ENERGIES, and the default zenith angles are given by constants.ANGLES_FOR_S_FLUXES.

    Parameters
    ----------
    primary_model : str in {"GSF", "HG", "GH", "ZS", "ZSP", "PL27"} or tuple, optional (default: "GSF")
        The primary flux model to use in MCEq. Options:
        GSF  = GlobalSplineFitBeta
        HG   = HillasGaisser2012 (H3a)
        GH   = GaisserHonda
        ZS   = Zatsepin-Sokolskaya (Default)
        ZSP  = Zatsepin-Sokolskaya (PAMELA)
        PL27 = SimplePowerlaw27
        Alternatively, this can be set with a tuple. For example: (pm.GaisserStanevTilav, "3-gen").

    interaction_model : str, optional (default: "SIBYLL-2.3c")
        The hadronic interaction model to use in MCEq. See the Tutorial or MCEq documentation for a list of options.

    atmosphere : {"CORSIKA", "MSIS00"}, optional (default: "CORSIKA")
        The atmospheric model to use in MCEq. For US Standard Atmosphere, use "CORSIKA". For seasonal variations, use "MSIS00".

    location : str, optional (default: "USStd")
        The name of the location for which to calculate the surface fluxes. See the Tutorial or MCEq documentation for a list of options.

    month : str, optional (default: None)
        The month for which to calculate the surface fluxes. For US Standard Atmosphere, use None. For seasonal variations, use the month name.

    output : bool, optional (default: taken from constants.get_output())
        If True, an output file will be created to store the results.

    file_name : str, optional (default: constructed from input parameters)
        The name of the file in which to store the results. If output is False or None, this is ignored.

    force : bool, optional (default: False)
        If True, force the creation of new directories if required.

    test: bool, optional (default: False)
        For use in the file test_s_fluxes.py. If True, this will calculate surface fluxes for only three angles.

    Returns
    -------
    s_fluxes : NumPy ndarray
        A two-dimensional array containing the surface fluxes. The shape will be (91, 20), and the fluxes will be in units of [(cm^2 s sr MeV)^-1].
    """

    # Import packages

    from MCEq.core import MCEqRun
    import mceq_config
    import crflux.models as pm

    # Check values

    constants._check_constants(force=force)

    if output is None:
        output = constants.get_output()

    assert atmosphere in [
        "CORSIKA",
        "MSIS00",
    ], 'atmosphere must be set to either "CORSIKA" or "MSIS00".'

    primary_models = {
        "GSF": (pm.GlobalSplineFitBeta, None),
        "HG": (pm.HillasGaisser2012, "H3a"),
        "GH": (pm.GaisserHonda, None),
        "ZS": (pm.ZatsepinSokolskaya, "default"),
        "ZSP": (pm.ZatsepinSokolskaya, "pamela"),
        "PL27": (pm.SimplePowerlaw27, None),
    }

    if isinstance(primary_model, str):

        assert (
            primary_model in primary_models
        ), "Set primary model not available. See the available options in the Tutorial at {0}.".format(
            "https://github.com/wjwoodley/mute/blob/main/docs/Tutorial.md#changing-the-primary-model"
        )

        primary_model_for_MCEq = primary_models[primary_model]
        pm_sname = primary_model

    elif isinstance(primary_model, tuple) and len(primary_model) == 2:

        primary_model_for_MCEq = primary_model
        pm_sname = primary_model[0](primary_model[1]).sname

    else:

        raise TypeError(
            "Primary model not set correctly. For an explanation, see the Tutorial at {0}.".format(
                "https://github.com/wjwoodley/mute/blob/main/docs/Tutorial.md#changing-the-primary-model"
            )
        )

    # Set the angles for use in MCEq

    angles = constants.ANGLES_FOR_S_FLUXES

    if test:

        angles = [0, 30, 60]

    # Set MCEq up

    mceq_config.enable_default_tracking = False

    if constants.get_verbose() > 1:

        print(
            "Calculating surface fluxes for {0} using {1} and {2}.".format(
                location, pm_sname, interaction_model
            )
        )

        mceq_run = MCEqRun(
            interaction_model=interaction_model,
            primary_model=primary_model_for_MCEq,
            theta_deg=0.0,
            **mceq_config.config,
        )
        mceq_run.set_density_model((atmosphere, (location, month)))

    else:

        from contextlib import redirect_stdout

        with open(os.devnull, "w") as suppress, redirect_stdout(suppress):

            mceq_run = MCEqRun(
                interaction_model=interaction_model,
                primary_model=primary_model_for_MCEq,
                theta_deg=0.0,
                **mceq_config.config,
            )
            mceq_run.set_density_model((atmosphere, (location, month)))

    # Calculate the surface fluxes
    # Zeroth index = Surface energy
    # First index  = Zenith angle

    s_fluxes = np.zeros((len(constants.ENERGIES), len(angles)))

    # Run MCEq
    # Convert the surface fluxes from default [GeV] to [MeV]

    for j in (
        tqdm(range(len(angles))) if constants.get_verbose() >= 1 else range(len(angles))
    ):

        mceq_run.set_theta_deg(angles[j])
        mceq_run.solve()

        s_fluxes[:, j] = (
            1e-3
            * (
                mceq_run.get_solution("total_mu+", mag=0)
                + mceq_run.get_solution("total_mu-", mag=0)
            )[:-30]
        )

    if constants.get_verbose() > 1:
        print("Finished calculating surface fluxes.")

    # Write the results to the file

    if output:

        constants._check_directory(
            os.path.join(constants.get_directory(), "surface"), force=force
        )

        if file_name == "" or not isinstance(file_name, str):

            file_name = os.path.join(
                constants.get_directory(),
                "surface",
                "Surface_Fluxes_{0}_{1}_{2}_{3}.txt".format(
                    location, str(month), interaction_model, pm_sname
                ),
            )

        file_out = open(file_name, "w")

        for i in range(len(constants.ENERGIES)):

            for j in range(len(angles)):

                file_out.write(
                    "{0:1.14f} {1:1.5f} {2:1.14e}\n".format(
                        constants.ENERGIES[i],
                        angles[j],
                        s_fluxes[i, j],
                    )
                )

        file_out.close()

        if constants.get_verbose() > 1:
            print(f"Surface fluxes written to {file_name}.")

    return s_fluxes


# Get surface fluxes


def load_s_fluxes_from_file(
    primary_model="GSF",
    interaction_model="SIBYLL-2.3c",
    atmosphere="CORSIKA",
    location="USStd",
    month=None,
    file_name="",
    force=False,
    test=False,
):

    """
    Retrieve a surface fluxes matrix in units of [(cm^2 s sr MeV)^-1] stored in data/surface based on the input parameters.

    If file_name is not given, this function searches for a file name that matches the set location, month, interaction model, and primary model. If the file does not exist, prompt the user to run calc_s_fluxes().

    Parameters
    ----------
    primary_model : str in {"GSF", "HG", "GH", "ZS", "ZSP", "PL27"} or tuple, optional (default: "GSF")
        The primary flux model to use in MCEq. Options:
        GSF  = GlobalSplineFitBeta
        HG   = HillasGaisser2012 (H3a)
        GH   = GaisserHonda
        ZS   = Zatsepin-Sokolskaya (Default)
        ZSP  = Zatsepin-Sokolskaya (PAMELA)
        PL27 = SimplePowerlaw27
        Alternatively, this can be set with a tuple. For example: (pm.GaisserStanevTilav, "3-gen").

    interaction_model : str, optional (default: "SIBYLL-2.3c")
        The hadronic interaction model to use in MCEq. See the Tutorial or MCEq documentation for a list of options.

    atmosphere : {"CORSIKA", "MSIS00"}, optional (default: "CORSIKA")
        The atmospheric model to use in MCEq. For US Standard Atmosphere, use "CORSIKA". For seasonal variations, use "MSIS00".

    location : str, optional (default: "USStd")
        The name of the location for which to calculate the surface fluxes. See the Tutorial or MCEq documentation for a list of options.

    month : str, optional (default: None)
        The month for which to calculate the surface fluxes. For US Standard Atmosphere, use None. For seasonal variations, use the month name.

    force : bool, optional (default: False)
        If True, force the calculation of a new surface flux matrix and the creation of new directories if required.

    test: bool, optional (default: False)
        For use in the file test_s_fluxes.py. If True, this will calculate surface fluxes for only three angles.

    Returns
    -------
    s_fluxes : NumPy ndarray
        A two-dimensional array containing the surface fluxes. The shape will be (91, 20), and the fluxes will be in units of [(cm^2 s sr MeV)^-1].
    """

    # Check values

    constants._check_constants(force=force)

    if isinstance(primary_model, str):

        pass

    elif isinstance(primary_model, tuple) and len(primary_model) == 2:

        raise TypeError(
            "The primary model must be set with a string. To set it with a tuple, run the surface.calc_s_fluxes() function with output set to True. For an explanation, see the Tutorial at {0}.".format(
                "https://github.com/wjwoodley/mute/blob/main/docs/Tutorial.md#changing-the-primary-model"
            )
        )

    else:

        raise TypeError(
            "Primary model not set correctly. For an explanation, see the Tutorial at {0}.".format(
                "https://github.com/wjwoodley/mute/blob/main/docs/Tutorial.md#changing-the-primary-model"
            )
        )

    # Set the angles

    angles = constants.ANGLES_FOR_S_FLUXES

    if test:

        angles = [0, 30, 60]

    # Define a function to run if there is no surface fluxes file

    def no_file(force):

        # If the file does not exist, ask the user if they want to run MCEq to create it

        if not force:

            answer = input(
                "No surface flux matrix currently exists for these models. Would you like to create one (y/n)?: "
            )

        if force or answer.lower() == "y":

            s_fluxes = calc_s_fluxes(
                primary_model,
                interaction_model,
                atmosphere,
                location,
                month,
                force=force,
                test=test,
            )

            return s_fluxes

        else:

            print("Surface fluxes not calculated.")

            return None

    # Construct a file name based on the user's inputs if one has not been provided

    if file_name == "" or not isinstance(file_name, str):

        file_name = os.path.join(
            constants.get_directory(),
            "surface",
            "Surface_Fluxes_{0}_{1}_{2}_{3}.txt".format(
                location, str(month), interaction_model, primary_model
            ),
        )

    # Check if the file exists

    if os.path.exists(file_name):

        if constants.get_verbose() > 1:
            print(
                "Loading surface fluxes for {0} using {1} and {2}.".format(
                    location, primary_model, interaction_model
                )
            )

        # Check that the file has the correct numbers of energies and zenith angles

        file = open(file_name, "r")
        n_lines = len(file.read().splitlines())
        file.close()

        # If so, read the surface fluxes in from it

        if n_lines == len(constants.ENERGIES) * len(angles):

            s_fluxes = np.reshape(
                np.loadtxt(file_name)[:, 2], (len(constants.ENERGIES), len(angles))
            )

            if constants.get_verbose() > 1:

                print("Loaded surface fluxes.")

            return s_fluxes

        # If the file does not have the correct numbers of energies and zenith angles, run the no_file() function

        else:

            return no_file(force=force)

    # If the file does not exist, run the no_file() function

    else:

        return no_file(force=force)


# Calculate the surface intensities


def calc_s_intensities(
    s_fluxes=None,
    primary_model="GSF",
    interaction_model="SIBYLL-2.3c",
    atmosphere="CORSIKA",
    location="USStd",
    month=None,
    output=None,
    file_name="",
    force=False,
):

    """
    Calculate surface intensities in units of [(cm^2 s sr)^-1] for default surface energy grid and surface flux zenith angles.

    The default surface energy grid is given by constants.ENERGIES, and the default zenith angles are given by constants.ANGLES_FOR_S_FLUXES.

    Parameters
    ----------
    s_fluxes : NumPy ndarray, optional (default: taken from surface.load_s_fluxes_from_file())
        A surface flux matrix of shape (91, 20).

    primary_model : str in {"GSF", "HG", "GH", "ZS", "ZSP", "PL27"} or tuple, optional (default: "GSF")
        The primary flux model to use in MCEq. Options:
        GSF  = GlobalSplineFitBeta
        HG   = HillasGaisser2012 (H3a)
        GH   = GaisserHonda
        ZS   = Zatsepin-Sokolskaya (Default)
        ZSP  = Zatsepin-Sokolskaya (PAMELA)
        PL27 = SimplePowerlaw27
        Alternatively, this can be set with a tuple. For example: (pm.GaisserStanevTilav, "3-gen").

    interaction_model : str, optional (default: "SIBYLL-2.3c")
        The hadronic interaction model to use in MCEq. See the Tutorial or MCEq documentation for a list of options.

    atmosphere : {"CORSIKA", "MSIS00"}, optional (default: "CORSIKA")
        The atmospheric model to use in MCEq. For US Standard Atmosphere, use "CORSIKA". For seasonal variations, use "MSIS00".

    location : str, optional (default: "USStd")
        The name of the location for which to calculate the surface fluxes. See the Tutorial or MCEq documentation for a list of options.

    month : str, optional (default: None)
        The month for which to calculate the surface fluxes. For US Standard Atmosphere, use None. For seasonal variations, use the month name.

    output : bool, optional (default: taken from constants.get_output())
        If True, an output file will be created to store the results.

    file_name : str, optional (default: constructed from input parameters)
        The name of the file in which to store the results. If output is False or None, this is ignored.

    force : bool, optional (default: False)
        If True, force the calculation of a new surface flux matrix and the creation of new directories if required.

    Returns
    -------
    s_intensities : NumPy ndarray
        A one-dimensional array containing the surface intensities. The length will be 20, and the intensities will be in units of [(cm^2 s sr)^-1].
    """

    # Check values

    constants._check_constants()

    if output is None:
        output = constants.get_output()

    # Get the surface flux matrix

    if constants.get_verbose() > 1:
        print("Calculating surface intensities.")

    if s_fluxes is None:

        s_fluxes = load_s_fluxes_from_file(
            primary_model, interaction_model, atmosphere, location, month, force=force
        )

    # Check that the surface flux matrix has been loaded properly

    if s_fluxes is None:

        raise Exception(
            "Surface intensities not calculated. The surface flux matrix was not provided or loaded correctly."
        )

    s_fluxes = np.atleast_2d(s_fluxes)

    # Calculate the surface intensities

    s_intensities = [
        scii.simpson(s_fluxes[:, j], constants.ENERGIES)
        for j in range(len(constants.ANGLES_FOR_S_FLUXES))
    ]

    if constants.get_verbose() > 1:
        print("Finished calculating surface intensities.")

    # Write the results to the file

    if output:

        constants._check_directory(
            os.path.join(constants.get_directory(), "surface"), force=force
        )

        if file_name == "" or not isinstance(file_name, str):

            file_name = os.path.join(
                constants.get_directory(),
                "surface",
                "Surface_Intensities_{0}_{1}_{2}_{3}.txt".format(
                    location, str(month), interaction_model, primary_model
                ),
            )

        file_out = open(file_name, "w")

        for j in range(len(constants.ANGLES_FOR_S_FLUXES)):

            file_out.write(
                "{0:1.5f} {1:1.14e}\n".format(
                    constants.ANGLES_FOR_S_FLUXES[j], s_intensities[j]
                )
            )

        file_out.close()

        if constants.get_verbose() > 1:
            print(f"Surface intensities written to {file_name}.")

    return s_intensities


# Calculate total surface fluxes


def calc_s_tot_flux(
    s_fluxes=None,
    primary_model="GSF",
    interaction_model="SIBYLL-2.3c",
    atmosphere="CORSIKA",
    location="USStd",
    month=None,
    force=False,
):

    """
    Calculate a total surface flux in units of [(cm^2 s)^-1] for default surface energy grid and surface flux zenith angles.

    The default surface energy grid is given by constants.ENERGIES, and the default zenith angles are given by constants.ANGLES_FOR_S_FLUXES.

    Parameters
    ----------
    s_fluxes : NumPy ndarray, optional (default: taken from surface.load_s_fluxes_from_file())
        A surface flux matrix of shape (91, 20).

    primary_model : str in {"GSF", "HG", "GH", "ZS", "ZSP", "PL27"} or tuple, optional (default: "GSF")
        The primary flux model to use in MCEq. Options:
        GSF  = GlobalSplineFitBeta
        HG   = HillasGaisser2012 (H3a)
        GH   = GaisserHonda
        ZS   = Zatsepin-Sokolskaya (Default)
        ZSP  = Zatsepin-Sokolskaya (PAMELA)
        PL27 = SimplePowerlaw27
        Alternatively, this can be set with a tuple. For example: (pm.GaisserStanevTilav, "3-gen").

    interaction_model : str, optional (default: "SIBYLL-2.3c")
        The hadronic interaction model to use in MCEq. See the Tutorial or MCEq documentation for a list of options.

    atmosphere : {"CORSIKA", "MSIS00"}, optional (default: "CORSIKA")
        The atmospheric model to use in MCEq. For US Standard Atmosphere, use "CORSIKA". For seasonal variations, use "MSIS00".

    location : str, optional (default: "USStd")
        The name of the location for which to calculate the surface fluxes. See the Tutorial or MCEq documentation for a list of options.

    month : str, optional (default: None)
        The month for which to calculate the surface fluxes. For US Standard Atmosphere, use None. For seasonal variations, use the month name.

    force : bool, optional (default: False)
        If True, force the calculation of new arrays or matrices and the creation of new directories if required.

    Returns
    -------
    s_tot_flux : float
        The total surface flux in units of [(cm^2 s)^-1].
    """

    # Check values

    constants._check_constants()

    # Calculate the surface intensities

    if constants.get_verbose() > 1:
        print("Calculating total surface flux.")

    s_intensities = calc_s_intensities(
        s_fluxes,
        primary_model,
        interaction_model,
        atmosphere,
        location,
        month,
        force=force,
    )

    if s_intensities is None:

        raise Exception(
            "Total surface flux not calculated. The surface intensities were not calculated properly."
        )

    # Calculate the total surface flux
    # Because constants.ANGLES_FOR_S_FLUXES goes (0..89), cos(angles) goes (1..0)
    # cos(constants.ANGLES_FOR_S_FLUXES) is decreasing, but scii.simpson() wants an increasing array
    # Therefore, integrate backwards, using [::-1] on the integrand and steps
    # Otherwise, the answer will be negative

    s_tot_flux = (
        2
        * np.pi
        * scii.simpson(
            s_intensities[::-1], np.cos(np.radians(constants.ANGLES_FOR_S_FLUXES[::-1]))
        )
    )

    if constants.get_verbose() > 1:
        print("Finished calculating total surface flux.")

    return s_tot_flux



import pytest
import sys

import numpy as np

import mute.constants as mtc
import mute.propagation as mtp

try:

    import proposal as pp

except ImportError:

    pass


@pytest.mark.skipif("proposal" not in sys.modules, reason="Requires proposal.")
def test_propagation():

    mtc.clear()

    mtc.set_n_muon(3)
    mtp._create_propagator(force=True)
    pp.RandomGenerator.get().set_seed(500)

    u_energy_calc = mtp._propagation_loop(
        mtc.ENERGIES[50], mtc.slant_depths[0], force=True
    )
    u_energy_read = [8088593.167141137, 7698826.729767787, 7568993.179976861]

    assert np.allclose(u_energy_calc, u_energy_read)
